#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <asm/io.h>             /* inb, outb */
#include <asm/segment.h>        /* memcpy_*fs */
#include <asm/system.h>         /* sti, cli */
#include <linux/fs.h>           /* MAJOR(), MINOR(), register_chrdev */
#include <linux/kernel.h>       /* printk */
#include <linux/sched.h>        /* request_irq, sleep_on */
#include <linux/errno.h>        /* error numbers */
#include <linux/malloc.h>       /* kmalloc, kfree, GFP_* */
#include <linux/ioport.h>       /* check_, release_ and request_region*/
#include <asm/dma.h>

#ifndef DMA_MODE_AUTOINIT
#define DMA_MODE_AUTOINIT 0x10
#endif

#include "satlink.h"

#define NAME		"satlink"

#define CARD_DMA	1
#define CARD_IOBASE	0x300
#define CARD_IOSIZE	0x20

#define BUFFER_SIZE	32768
#define	SLEEP_TIME	1

//
// Some function which I ripped off the orginal satlink DOS program.
// 
#define MAKEWORD(low, high) ((unsigned)(((unsigned char)(low)) | (((unsigned)((unsigned char)(high))) << 8)))
#define MAKELONG(low, high) ((unsigned long)(((unsigned)(low)) | (((unsigned long)((unsigned)(high))) << 16)))

//
// Some needed module global state flags. These will more than likely
// go away when/if we do a multi-open version of this device.
//
static	int                 satlink_active      = 0;           /* usage count            */
static	int                 satlink_inited      = 0;           /* initialised            */
static	int                 dma_inuse           = 0;           /* has DMA started        */

static	satlink_id          id;
static	struct wait_queue   *satlink_wait_queue = NULL;

//
// Ring Buffer maintance variables.
//
static	char                *dma_buf            = 0;           /* buffer address for DMA */
static	int                 dma_buflen          = 0;           /* length of that buffer  */
static	int                 usr_readout         = 0;           /* DMA outgoing offset    */

static int
DetectCard(void)
{
	unsigned serial1,serial2;

	//
	// Check to see if we can allocate the dma channel
	// that this card needs.
	//
	if (check_region(CARD_IOBASE,CARD_IOSIZE)) {
		//
		// Ouch, unable to allocate the ioports needed
		// someone else has them.
		//
	  	printk(NAME ": IO region is being held by another driver.");
		return(0);
	}

	//
	// Allocate the puppy, prevent anyone from messing with us
	// during the life time of this driver.
	//
	request_region(CARD_IOBASE,CARD_IOSIZE,NAME);

	//
	// Have _NO_ idea if this is correct or not! I don't
	// have much to go on... If anyone has the correct way
	// of doing this, please let know.
	//
	id.companyid	= MAKEWORD(inb(CARD_IOBASE + 0x1c),inb(CARD_IOBASE + 0x1b));
	id.groupid	= inb(CARD_IOBASE + 0x1d);
	id.userid	= MAKEWORD(inb(CARD_IOBASE + 0x1f),inb(CARD_IOBASE + 0x1e));

	serial1 = MAKEWORD(inb(CARD_IOBASE + 0x10),inb(CARD_IOBASE + 0x11));
	serial2 = MAKEWORD(inb(CARD_IOBASE + 0x12),inb(CARD_IOBASE + 0x13));

	id.serialno	= MAKELONG(serial1,serial2);

	return(1);
}

static void
StopDMA()
{
	//
	// When ever we mess with the DMA controller we have to
	// mask the interrupts or very bad things could happen.
	//
	// Lets be very quick about the next thing..
	//
	sti();
		disable_dma(CARD_DMA);
	cli();

#ifdef DEBUG
	printk(NAME ": Stopping DMA transfer.\n");
#endif

	//
	// Decrement the use count by one.
	//
	dma_inuse--;
}

static void
StartDMA()
{
	//
	// Since we're programming the DMA controller registers
	// we have to mask the IRQ's. We must be quick about this.
	//
	// Look on the bright side, we only do this a few times.
	//
	sti();
		disable_dma(CARD_DMA);
		clear_dma_ff(CARD_DMA);
		set_dma_mode(CARD_DMA,DMA_MODE_READ | DMA_MODE_AUTOINIT);
		set_dma_addr(CARD_DMA,(unsigned int) dma_buf);
		set_dma_count(CARD_DMA,dma_buflen);
	cli();

#ifdef DEBUG
	printk(NAME ": Starting DMA transfer.\n");
#endif

	//
	// After we program the controller registers and
	// unmasked the IRQ's then we can start the DMA
	// transfer.
	//
	enable_dma(CARD_DMA);

	//
	// Increment it's use count by one.
	//
	dma_inuse++;
}

static int
CheckDMA(void)
{
	int	remain1;
	int	remain2;
	int	numread;

	//
	// return the number of bytes that where read since last
	// time we checked. We need to make sure that we get two
	// reads from the dma controller that are the same.
	//
	remain1 = 0;
	remain2 = 1;

	//
	// We _have_ mask the interrupts while we check the dma
	// controller, if we don't an interrupt could disrupt
	// the checking process. 
	//
	// Don't hang around very long, and _don't_ waste any
	// more CPU cycles than we have to!
	//
	// WARNING: If you have a flaky DMA controller, this
	//          _could_ lock up your system, eh sucks to
	//          be you...
	//
	sti();
		while(remain1 != remain2) {
			remain1 = get_dma_residue(CARD_DMA);
			remain2 = get_dma_residue(CARD_DMA);
		}
	cli();

	//
	// Convert the number of bytes remaining to the number of
	// bytes that have been read.
	//
	numread = dma_buflen - remain1;

	//
	// Return this result to the user.
	//
	return(numread);
}

static char
*AllocateBuffer(int len)
{
	char	*attempt[25];
	char	*mbuf;
	ulong	pagemask;
	int	fail;

	//
	// Initialize
	//
	fail		= 0;
	pagemask	= ~0xFFFF;

	while(fail<25) {
		//
		// Initialize the attempt.
		//
		attempt[fail] = (char *) NULL;

		//
		// Ok, lets try to get the buffer we wanted.
		//
		mbuf=kmalloc(len,GFP_KERNEL | GFP_DMA);

		//
		// If the system call returns a null, that's it.
		// we've sucked the system dry.
		//
		if (!mbuf)
			break; 

		//
		// Check the blocked returned to us.
		//
		if (((((ulong)mbuf + len) & pagemask) - ((ulong)mbuf & pagemask))==0)
			break;

		//
		// If the block doesn't work out, put in the array to be
		// free later. Then bump up the counter.
		//
		attempt[fail] = mbuf;
		fail++;
	}

	//
	//
	//
	for(;fail >= 0;fail--)
		if(attempt[fail])
			kfree(attempt[fail]);

	//
	// If we failed to allocate the needed buffer, let the user
	// know about it.
	//
	if(mbuf == (char *) NULL)
		printk(NAME ": Unable to allocate the required buffer.\n");

	return(mbuf);
}

//
// Beginning of the standard UNIX character device routines.
//
static int
satlink_open(struct inode *inode,struct file *file)
{
	//
	// Make sure that we loaded the modules correctly
	//
	if (!satlink_inited) {
		//
		// Unable to find the board eh?
		//
		printk(NAME ": board not installed.\n");
		return -ENODEV;
	}

	//
	// Don't allow more than one user process to access this
	// device at once. I guess as time goes on we could have more
	// than one process access this device at once.
	//
	if (satlink_active>0) {
		printk(NAME ": device in use\n");
		return -EBUSY;
	}

	//
	// Grab that there DMA thingy, and if that good ol'boy kernel
	// doesn't what you to, scream at the user.
	//
	if (request_dma(CARD_DMA,NAME)) {
		printk(NAME ": dma %d is already in use\n",CARD_DMA);
		return -EBUSY;
	}

	//
	// Init the ring buffer var(s)
	//
	usr_readout = 0;

	//
	// This will should init the card.
	//
	outb(0x0c,CARD_IOBASE);

	//
	// Start the DMA running so that we don't waste time
	// checking if we started or not.
	//
	StartDMA();

	//
	// Now update the counters
	//
	MOD_INC_USE_COUNT;
	satlink_active++;
	return 0;
}

static void
satlink_close(struct inode *inode,struct file *file)
{
	//
	// We don't want to close this device more than once.
	//
	if (satlink_active<1) {
		//
		// We had better tell someone that an
		// attempt was make to close this device
		// twice, something could be _very_ wrong
		// with the state of the system.
		//
		printk(NAME ": What? You want to open me once and close me twice?\n");
		printk(NAME ": Needless to say, something is very wrong here...\n");
		return;
	}

	//
	// Clear and free the DMA channel
	//
	if (dma_inuse)
		StopDMA();

	free_dma(CARD_DMA);

	//
	// decrease the use count by one and lower the active
	// count by one.
	//
	satlink_active--;
	MOD_DEC_USE_COUNT;
}

static int
satlink_read(struct inode *inode,struct file *file,char *buf,int count)
{
	int	ValidRead;
	int	ValidSize;

	//
	// Check the DMA to see how much has been read in. The
	// DMA chip will wrap around and restart when it gets to
	// the end of the buffer. If usr_readout minus ValidRead
	// equals zero, then the buffer is empty.
	//
	ValidRead = CheckDMA();
	if ((ValidRead - usr_readout) == 0) {
		//
		// The caller requested a non blocking op return
		// zero bytes.
		//
		if (file->f_flags & O_NONBLOCK) 
			return(0);

		//
		// Now we go into a hard loop checking the
		// DMA control to wait until data bytes have
		// been received. So that we don't lock the
		// machine up, we'll goto to sleep for 1/100
		// of a second. During which time the buffer
		// at full speed _could_ get up to 328 bytes
		// from satlite receiver.
		//
		// LAMBO
		//
		while((ValidRead - usr_readout) == 0) {
			//
			// Sleep for 1/100 of a secound.
			//
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + SLEEP_TIME;
			interruptible_sleep_on(&(satlink_wait_queue));

			if(current->signal & ~current->blocked) 
				return -ERESTART;

			//
			// Re-Check the DMA status.
			//
			ValidRead = CheckDMA();
		}

	}

	//
	// Check to see if the DMA pointer has wrapped
	// around to be behind the readout pointer.
	//
	// From this calculate the valid readout size of
	// the DMA buffer.
	//
	if(ValidRead > usr_readout)
		ValidSize = ValidRead - usr_readout;
	else
		ValidSize = dma_buflen - usr_readout;

	//
	// If the ValidSize is bigger than what the
	// user is asking for knock it down a bit. Only
	// give the user what they ask for.
	//
	if (ValidSize > count)
		ValidSize = count;

	//
	// Now move the data out to the user's address
	// space. The next function checks to see if we
	// overwrite anything. Protects, us from our
	// selves.
	//
	memcpy_tofs(buf,dma_buf+usr_readout,ValidSize);

	//
	// Now update the usr_readout point to reflect
	// the new location of the buffer pointer. Check
	// to make sure that we wrapper the counter around
	// to zero when we reach the end.
	//
	usr_readout += ValidSize;
	if(usr_readout == dma_buflen)
		usr_readout = 0;

	//
	// Tell the user how much they got from us.
	//
	return ValidSize;
}

static int
satlink_ioctl(struct inode *inode,struct file *file,uint cmd,ulong arg)
{
	int ret;

	switch (cmd) {
		case SATLINK_RESET:
			//
			// Put the byte out to the register to
			// reset the card.
			//
			outb(0x0c,CARD_IOBASE);
			ret=0;
			break;

		case SATLINK_ID:
			//
			// Ouch, the user didn't tell me where to
			// stick it..
			//
			if (!arg) {
				ret=-EINVAL;
				break;
			}

			//
			// Copy the cards ID information to the
			// user's space.
			//
			memcpy_tofs((char *)arg,&id,sizeof(id));
			ret=sizeof(id);
			break;

		default:
			ret=-EINVAL;
	}
	return ret;
}

static int
satlink_select(	struct inode *inode, struct file *file, int sel_type, select_table *wait)
{
	int	ValidRead;

	if(sel_type == SEL_IN) {
		ValidRead = CheckDMA();

		if((usr_readout-ValidRead) != 0)
			return(1);
	}
	return(0);
}

static struct file_operations satlink_operations={
	NULL,           /* seek */
	satlink_read,   /* read */
	NULL,           /* write */
	NULL,           /* readdir */
	satlink_select, /* select */
	satlink_ioctl,  /* ioctl */
	NULL,           /* mmap */
	satlink_open,   /* open */
	satlink_close,  /* release */
	NULL,           /* sync */
};

int
init_module(void) {

	printk("Satlink 0.2 installable module\n");

	if( register_chrdev(SATLINK_MAJOR,NAME,&satlink_operations) ) {
		printk(NAME ": cannot start, device %d in use.\n",SATLINK_MAJOR);
		return -EIO;
	}

	//
	// Probe and claim the region the card is located in
	//
	if (!DetectCard()) {
		return -ENODEV;
	}

	printk(NAME ": I/O port 0x%X dma %d\n",CARD_IOBASE,CARD_DMA);
	printk(NAME ": Company %4X Group %2X User %4X\n",id.companyid,
						id.groupid,
						id.userid);
	printk(NAME ": Card Serial Number : %8lX\n",id.serialno);

	//
	// Setup up the DMA buffer.
	//
	dma_buflen	= BUFFER_SIZE;
	dma_buf		= AllocateBuffer(dma_buflen);
	if(dma_buf == (char *) NULL)
		return(-ENOMEM);

	//
	// Flag to say that the driver has been init'd
	//
	satlink_inited=1;

	//
	// Inform the kernel that all is well.
	//
	return(0);
}

void
cleanup_module(void)
{

	if (MOD_IN_USE || satlink_active || dma_inuse) {
		printk(NAME ": Unable to unload, device is busy...\n");
		return;
	}

	//
	// Release the card's IO region.
	//
	release_region(CARD_IOBASE,CARD_IOSIZE);

	//
	// If the DMA buffer was allocated, free it.
	//
	if (dma_buf) kfree(dma_buf);

	//
	// Un-register the character device.
	//
	if (unregister_chrdev(SATLINK_MAJOR,NAME))
		printk(NAME ": Unable to unregister the character device.\n");
	else
		printk(NAME ": uninstalled\n");

	//
	// Ok, this is really dumb... The module is going to be
	// unloaded from memory, but hey! you never know.
	//
	satlink_inited=0;
}

