
/*
 * Dev32.satlink
 *
 * Qnx4 Realtime OS,
 *
 * Jean-Claude MICHOT (jcmichot@teaser.fr)
 * France-Teaser, Jun 1997
 *
 * This driver is designed to receive "Planet Connect" Satellite Feed
 *
 * Planet Connect:   http://www.planetc.com
 * France-Teaser:    http://www.teaser.fr
 *
 * This driver is based on the sample device driver for Qnx4.
 *
 * One point worth mentioning is that "most" hardware drivers
 * will provide input into the Device subsystem directly from
 * the input interrupt handler.
 * This Dev32.satlink driver use a fixed 50 ms timer for
 * process input data from DMA mpeg hardware board.
 *
 * On the other hand, it is NOT uncommon for a "real" driver to
 * output data at process time as is the case in this example.
 * (for example Dev.con, Dev.win, Dev.par all output data at
 * process time).
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/sched.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/psinfo.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

//#include <sys/_devdrvr.h>
#include "_devdrvr.h"

extern int far Kick (struct output_buffer far *obuf);
extern int far Stty (struct termios far *tios); 
extern int far Ctrl (union ioctl_entry far *ioc); 

extern int sys_msg( void * );

union _private_msg {
	msg_t		type;
	msg_t		status;
	char		filer[100];
	} msg;

int idle = 1;
int more = 0;

pid_t out_proxy;
pid_t in_proxy;
pid_t proxy;

unsigned isize = 56*1024L;
unsigned osize = 56*1024L;
unsigned csize = 1024L;

#define MAX_DEVICES 10

struct driver_ctrl drv_ctrl;
struct device_ctrl ctrl[MAX_DEVICES];
int dev_handle;
int num_devices = 1;
 /* found in <sys/_devdrvr.h> */
unsigned driver_options = DRVR_NO_SUFFIX; 
	// Suppress the numeric suffix if only	1

char *prefix = "satlink";
static int debug = 0;

main(argc, argv)
int argc; char *argv[];
	{
	unsigned c;
	int i;
	pid_t pid;
	int t;
	int nbytes;
	extern void terminate();

	setscheduler(0, SCHED_FIFO, 15);

	/*
	 * Allow only SIGTERM, which we catch and cause
	 * graceful termination, all others are ignored
	 */
	for( i = _SIGMIN; i <= _SIGMAX; ++i)
		if( i != SIGTERM)
			signal( i, SIG_IGN );

	signal( SIGTERM, &terminate );

	/*
	 * Scan arguments
	 */
	while ((c=getopt(argc,argv,"C:I:O:N:n:D")) != -1) {
		switch (c) {

		case 'C':
			csize = strtol(optarg,NULL,0);
			break;

		case 'I':
			isize = strtol(optarg,NULL,0);
			break;

		case 'O':
			osize = strtol(optarg,NULL,0);
			break;

		case 'N':
			prefix = optarg;
			break;

		case 'n':
			num_devices = strtol(optarg,NULL,0);
			if(num_devices > MAX_DEVICES)
				num_devices = MAX_DEVICES;
			break;

		case 'D':
			debug++;
			break;

			}
		}

	/*
	 * Register ourselves with the DEVICE Manager
	 */

	if((dev_handle = dev_drvr_register( "satlink", prefix, num_devices,
					&Kick, &Stty, &Ctrl, &drv_ctrl, driver_options)) == -1) {
		exit(ENFILE);
		}

	/*
	 * Allocate the resources for this device
	 */

	for(i = 0; i < num_devices; ++i) {
		ctrl[i].isize = isize;
		ctrl[i].osize = osize;
		ctrl[i].csize = csize;
		}

	if(dev_drvr_mount(dev_handle, drv_ctrl.base_tty, num_devices, &ctrl[0])) {
		dev_drvr_shutdown( dev_handle );
		exit(ENOMEM);
		}

	out_proxy = qnx_proxy_attach(0, 0, 0, -1);
	in_proxy = qnx_proxy_attach(0, 0, 0, -1);

	/*
	 * Close off stdin,stdout,stderr since this process
	 * is usually started in sysinit as a background process,
	 * by the shell, which typically sets stdin to /dev/null
	 * (on the machine we booted from), which ties up resources
	 */

#if 0 // I want see the debug printf
	close(0);
	close(1);
	close(2);
#endif

	/*
	 * Tell Proc that we will respond to "sin ver" messages
	 */
	qnx_pflags( _PPF_SERVER, _PPF_SERVER, 0, 0 );

	/*
	 * Initialize termios and queue parameters
	 * to whatever defaults make sense for this class of device
	 */
	
	for( i = 0; i < num_devices; ++i ) {
		/* struct input_buffer far *ibuf; */
		struct output_buffer far *obuf;
		struct termios far *tios;

		obuf = ctrl[i].obuf;
		//obuf->mode |= OUTPUT_XNL;

		tios = ctrl[i].termios;

		tios->c_ispeed = tios->c_ospeed = 9600;

		// Define a RAW device for satellite receiver

		tios->c_cflag = (CREAD);
		tios->c_oflag = 0; // (OPOST);
		tios->c_iflag = 0; // (ICRNL|IXON|BRKINT);
		tios->c_lflag = IEXTEN; // (ICANON|ISIG|ECHO|ECHOE|ECHOK|IEXTEN);

		tios->c_cc[VEOF]	= //0x04;
		tios->c_cc[VEOL]	= //0x00;
		tios->c_cc[VERASE]	= //0x7F;
		tios->c_cc[VINTR]	= //0x03;
		tios->c_cc[VKILL]	= //0x18;
		tios->c_cc[VQUIT]	= //0x00;
		tios->c_cc[VSUSP]	= //0x00;
		tios->c_cc[VSTART]	= //0x11;
		tios->c_cc[VSTOP]	= //0x13;
								0;

		tios->c_cc[VMIN]	= 0x01;
		tios->c_cc[VTIME]	= 0x00;
		}

	/*
	 * Initialize hardware
	 */

	init_driver();


	/*
	 * Let DEV call us now
	 */
	for(i = 0; i < num_devices; ++i) {
		if(dev_drvr_arm(ctrl[i].tty)) {
			dev_drvr_shutdown( dev_handle );
			exit(ENODEV);
			}
		}


	/*
	 * Main loop
	 */
	
	for(;;) {
	
		pid = Receive( 0, &msg, sizeof(msg) );

		//printf("Received from PID %ld\n", pid );

		if(pid == out_proxy) {
			idle = 1;
			/*
			 * Output data now available
			 */
			flush_output();
			continue;
			}

		if(pid == in_proxy) {
			/*
			 * Input data is available
			 * NOTE: that this would "normally" be taken care of
			 *       by an interrupt handler
			 */
			input_ready();
			continue;
			}

		/*
		 * Otherwise, process a message from a process
		 */

		t = msg.type;
		msg.status = ENOSYS;
		nbytes = sizeof(msg.status);

		switch(t) {

		case 0:
			nbytes = sys_msg(&msg);
			break;

		default:
			break;
			}

		if(nbytes)
			Reply(pid, &msg, nbytes);

		continue;		
		}
	}


void terminate( int signal )
	{
	if(signal) {
		/*
		 * Shutdown by signal
		 */
		dev_drvr_shutdown( dev_handle );
		/*
		 * We should give back any GDT's, etc. that
		 * the system doesn't release for us
		 */
		exit_driver();
		}
	exit(0);
	}


/*
 * Process all input data in all devices
 * Only return when all input is exhausted.
 * Rely on a subsequent Interrupt (which triggers 'in_proxy') to wake us up.
 */

int input_ready() {
	pid_t proxy;
	int i, n;

	char buf[8192];

	proxy = 0;

	/*
	 * Scan all devices looking for available data
	 */
	for( i = 0; i < num_devices; ++i ) {
		while( (n = get_data(i, &buf[0], sizeof(buf))) > 0 ) {
			if(drv_ctrl.ttim_offset != 0) {
				/*
				 * Use multiple character input routine if supported by Dev
				 */
				pid_t (far *ttimcall) (int tty, int len, char far *data, unsigned ds);
				ttimcall = MK_FP( FP_SEG(drv_ctrl.tti), drv_ctrl.ttim_offset );
				proxy |= (*ttimcall) (ctrl[i].tty, n, &buf[0], drv_ctrl.tti_ds);
				}
			else {
				/*
				 * Must inject character at a time for older Dev's
				 */
				int j;
				for(j = 0; j < n; ++j)
					proxy |= (*drv_ctrl.tti) (buf[j], ctrl[i].tty, drv_ctrl.tti_ds);
				}
			}
		}

	/*
	 * "wake up" Dev if ANYTHING of interest happened
	 */
	if(proxy)
		Trigger(proxy);

	return(0);
	}

/*
 * This function will print forever, only returning
 * when all output buffers are drained
 * Rely on a subsequent Kick() call to wake us up again
 *
 * NOTE: that the memory semaphores 'idle' and 'more' are updated
 *       by the far-call stubs (ie: Kick) to make this work efficiently
 */

flush_output() {
	struct output_buffer far *obuf;
	char far *tail;
	char far *end;
	char far *begin;
	int i, n, size;
	int nothing, unit;
	int expand_newlines;

	for(;;) {
		/*
		 * Scan every output queue, looking for something to do
		 */
		nothing = 0;
		more = 0;
		for(unit = 0; unit < num_devices; ++unit) {

			obuf = ctrl[unit].obuf;

			/*
			 * Process flow control output functions first
			 */

			if(obuf->mode & OUTPUT_HFLOW_OFF) {
				obuf->mode &= ~OUTPUT_HFLOW_OFF;
				drop_hw(unit);
				}
	
			if(obuf->mode & OUTPUT_HFLOW_ON) {
				obuf->mode &= ~OUTPUT_HFLOW_ON;
				raise_hw(unit);
				}

			if(obuf->mode & OUTPUT_XOFF) {
				obuf->mode &= ~OUTPUT_XOFF;
				/*
				 * Disable input by sending XOFF
		   	 	 */
				put_data(unit, (char far *) "\023", 1, 0);
				}
	
			if(obuf->mode & OUTPUT_XON) {
				obuf->mode &= ~OUTPUT_XON;
				/*
				 * Re-enable input by sending XON
			 	 */
				put_data(unit, (char far *) "\021", 1, 0);
				}
	
			/*
			 * Skip if no data characters available
			 */

			_disable();
			if(obuf->size == 0  ||
					(obuf->mode & (OUTPUT_STOPPED|OUTPUT_STOPPED_HW)) != 0) {
				_enable();
				++nothing;
				continue;
				}
			_enable();
	
			if(obuf->discard) {
				/*
				 * Discard some (or all) data instead of displaying it
				 */
				n = obuf->discard;
				_disable();
				obuf->discard -= n;
				obuf->size -= n;
				_enable();
				i = &obuf->buffer[obuf->length] - obuf->tail;
				if(i > n)
					obuf->tail += n;
				else
					obuf->tail = &obuf->buffer[n-i];
				}

			/*
			 * Extract the data from the buffer
			 */

			tail = obuf->tail;
			end = &obuf->buffer[obuf->length];
			begin = &obuf->buffer[0];
			size = obuf->size;

			expand_newlines = obuf->mode & OUTPUT_XNL;

			/*
			 * Calculate number of chars before wrap-around point
			 */
			i = end - tail;
			if(i >= size) { /* No wrap around */
				n = put_data(unit, tail, size, expand_newlines);
				}
			else {
				n = put_data(unit, tail, i, expand_newlines);
				n += put_data(unit, begin, size - i, expand_newlines);
				}

			/*
			 * Update the Queue pointers, now that the
			 * 'n' bytes of data has been extracted
			 */

			_disable();

			if(i > n) /* No wrap around */
				obuf->tail = tail + n;
			else
				obuf->tail = begin + (n-i);
			obuf->size -= n;
			/*
			 * Compensate for any flushed data which
			 * was in the process of being transmitted
			 */
			if(obuf->discard > n)
				obuf->discard -= n;
			else
				obuf->discard = 0;

			_enable();
	
			/*
			 * If size of buffer drops below notification threshold,
			 * inform DEV of the new information
			 */
			_disable();
			if(obuf->size < obuf->bcount) {
				obuf->bcount = 0;
				obuf->action |= ACT_OUTPUT_READY;
				_enable();
				Trigger(drv_ctrl.hw_pid);
				}
			else
				_enable();
			}
	
		/*
		 */
		_disable();
		if(nothing >= num_devices  &&  more == 0) {
			idle = 1;
			_enable();
			return;
			}
		else
			_enable();
		}
	}



/*******************************************************************

 Hardware specific functions

 [add "useful" stuff here]

 ******************************************************************/

volatile unsigned counter;
int hint_id;

#pragma off( check_stack );
pid_t far handler()
{
	//return( in_proxy );

	if ( (++counter%2)==0 )
		return( in_proxy );

	return(0);
}
#pragma on( check_stack );


int init_driver()
	{
	/* Your stuff here */
	satlink_init();

	if (  (hint_id=qnx_hint_attach( 0, &handler, FP_SEG(&counter) ))==-1 )
		{
		printf("Unable to attach interrupt\n");
		fflush( stdout );
		}

	return(0);
	}



int exit_driver()
	{

	/* Your stuff here */
	free_dma_buf( 0 );
	qnx_hint_detach( hint_id );

	return(0);
	}



get_data(int unit, char *buf, int nbytes)
	{
	/*
	 * Get (at most) 'nbytes' of data from 'unit' (into 'buf').
	 * Return #bytes read.
	 */
	int n = 0;

	unit = unit; /* keep compiler happy */

	/* Your stuff here */
	n = satlink_read( buf, nbytes );

	return(n);
	}



put_data(int unit, char far *buf, int nbytes, int expand_newlines)
	{
	/*
	 * Output 'nbytes' of data (in 'buf') to 'unit'.
	 * Return #bytes written (before expansion).
	 */

	/* Your stuff here */
	return(nbytes);
	}



void drop_hw(int unit) {
	/*
	 * Disable input by dropping Hardware handshaking
 	 */
	unit = unit; /* keep compiler happy */
	/* Your stuff here */
	}



void raise_hw(int unit) {
	/*
	 * Re-enable input by raising Hardware handshaking
 	 */
	unit = unit; /* keep compiler happy */
	/* Your stuff here */
	}

/*------------------------------------------------------------*/
/*- Qnx4 SATLINK low level driver ----------------------------*/
/*------------------------------------------------------------*/
/*- Ported from Linux driver by jcmichot@teaser.fr -----------*/
/*------------------------------------------------------------*/

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/inline.h>

#include <sys/kernel.h>
#include <sys/seginfo.h>
#include <sys/proc_msg.h>
#include <errno.h>
#include <string.h>
#include <i86.h>
#include <sys/seginfo.h>

extern void far *alloc_dma_buf( long nbytes );

static int selector = -1;
static long dma_addr;

typedef unsigned long ulong;

/* > satlink.h */
/* Daniel F. Smith, 1994-6 */

#include <sys/ioctl.h>
/* formatted ID structure from card ROM */
typedef struct {
	unsigned short companyid;
	unsigned char groupid;
	unsigned short userid;
	unsigned long serialno;
	} satlink_id;

#define SATLINK_MAJOR 60
/* Using magic code 's' for Satlink */
#define SATLINK_RESET           _IO ('s', 0)
#define SATLINK_ID              _IOR('s', 1, satlink_id)
#define SATLINK_BUFSIZE         _IOW('s', 2, int)

#pragma aux cli     = "pushfd" "cli" "pop eax" __parm [] __modify [eax];
#pragma aux sti     = "pushfd" "sti" "pop eax" __parm [] __modify [eax];

/*
 * NOTES about DMA transfers:
 *
 *  controller 1: channels 0-3, byte operations, ports 00-1F
 *  controller 2: channels 4-7, word operations, ports C0-DF
 *
 *  - ALL registers are 8 bits only, regardless of transfer size
 *  - channel 4 is not used - cascades 1 into 2.
 *  - channels 0-3 are byte - addresses/counts are for physical bytes
 *  - channels 5-7 are word - addresses/counts are for physical words
 *  - transfers must not cross physical 64K (0-3) or 128K (5-7) boundaries
 *  - transfer count loaded to registers is 1 less than actual count
 *  - controller 2 offsets are all even (2x offsets for controller 1)
 *  - page registers for 5-7 don't use data bit 0, represent 128K pages
 *  - page registers for 0-3 use bit 0, represent 64K pages
 *
 * DMA transfers are limited to the lower 16MB of _physical_ memory.  
 * Note that addresses loaded into registers must be _physical_ addresses,
 * not logical addresses (which may differ if paging is active).
 *
 *  Address mapping for channels 0-3:
 *
 *   A23 ... A16 A15 ... A8  A7 ... A0    (Physical addresses)
 *    |  ...  |   |  ... |   |  ... |
 *    |  ...  |   |  ... |   |  ... |
 *    |  ...  |   |  ... |   |  ... |
 *   P7  ...  P0  A7 ... A0  A7 ... A0   
 * |    Page    | Addr MSB | Addr LSB |   (DMA registers)
 *
 *  Address mapping for channels 5-7:
 *
 *   A23 ... A17 A16 A15 ... A9 A8 A7 ... A1 A0    (Physical addresses)
 *    |  ...  |   \   \   ... \  \  \  ... \  \
 *    |  ...  |    \   \   ... \  \  \  ... \  (not used)
 *    |  ...  |     \   \   ... \  \  \  ... \
 *   P7  ...  P1 (0) A7 A6  ... A0 A7 A6 ... A0   
 * |      Page      |  Addr MSB   |  Addr LSB  |   (DMA registers)
 *
 * Again, channels 5-7 transfer _physical_ words (16 bits), so addresses
 * and counts _must_ be word-aligned (the lowest address bit is _ignored_ at
 * the hardware level, so odd-byte transfers aren't possible).
 *
 * Transfer count (_not # bytes_) is limited to 64K, represented as actual
 * count - 1 : 64K => 0xFFFF, 1 => 0x0000.  Thus, count is always 1 or more,
 * and up to 128K bytes may be transferred on channels 5-7 in one operation. 
 *
 */

#define MAX_DMA_CHANNELS	8

/* The maximum address that we can perform a DMA transfer to on this platform */
#define MAX_DMA_ADDRESS      (PAGE_OFFSET+0x1000000)

/* 8237 DMA controllers */
#define IO_DMA1_BASE	0x00	/* 8 bit slave DMA, channels 0..3 */
#define IO_DMA2_BASE	0xC0	/* 16 bit master DMA, ch 4(=slave input)..7 */

/* DMA controller registers */
#define DMA1_CMD_REG        0x08    /* command register (w) */
#define DMA1_STAT_REG       0x08    /* status register (r) */
#define DMA1_REQ_REG        0x09    /* request register (w) */
#define DMA1_MASK_REG       0x0A    /* single-channel mask (w) */
#define DMA1_MODE_REG       0x0B    /* mode register (w) */
#define DMA1_CLEAR_FF_REG   0x0C    /* clear pointer flip-flop (w) */
#define DMA1_TEMP_REG       0x0D    /* Temporary Register (r) */
#define DMA1_RESET_REG      0x0D    /* Master Clear (w) */
#define DMA1_CLR_MASK_REG   0x0E    /* Clear Mask */
#define DMA1_MASK_ALL_REG   0x0F    /* all-channels mask (w) */

#define DMA2_CMD_REG        0xD0    /* command register (w) */
#define DMA2_STAT_REG       0xD0    /* status register (r) */
#define DMA2_REQ_REG        0xD2    /* request register (w) */
#define DMA2_MASK_REG       0xD4    /* single-channel mask (w) */
#define DMA2_MODE_REG       0xD6    /* mode register (w) */
#define DMA2_CLEAR_FF_REG   0xD8    /* clear pointer flip-flop (w) */
#define DMA2_TEMP_REG       0xDA    /* Temporary Register (r) */
#define DMA2_RESET_REG      0xDA    /* Master Clear (w) */
#define DMA2_CLR_MASK_REG   0xDC    /* Clear Mask */
#define DMA2_MASK_ALL_REG   0xDE    /* all-channels mask (w) */

#define DMA_ADDR_0              0x00    /* DMA address registers */
#define DMA_ADDR_1              0x02
#define DMA_ADDR_2              0x04
#define DMA_ADDR_3              0x06
#define DMA_ADDR_4              0xC0
#define DMA_ADDR_5              0xC4
#define DMA_ADDR_6              0xC8
#define DMA_ADDR_7              0xCC

#define DMA_CNT_0               0x01    /* DMA count registers */
#define DMA_CNT_1               0x03
#define DMA_CNT_2               0x05
#define DMA_CNT_3               0x07
#define DMA_CNT_4               0xC2
#define DMA_CNT_5               0xC6
#define DMA_CNT_6               0xCA
#define DMA_CNT_7               0xCE

#define DMA_PAGE_0              0x87    /* DMA page registers */
#define DMA_PAGE_1              0x83
#define DMA_PAGE_2              0x81
#define DMA_PAGE_3              0x82
#define DMA_PAGE_5              0x8B
#define DMA_PAGE_6              0x89
#define DMA_PAGE_7              0x8A

#define DMA_MODE_READ    0x44 /* I/O to memory, no autoinit, increment, single mode */
#define DMA_MODE_WRITE   0x48 /* memory to I/O, no autoinit, increment, single mode */
#define DMA_MODE_CASCADE 0xC0 /* pass thru DREQ->HRQ, DACK<-HLDA only */


static int dma_inb( val )
{
    return( inb( val ) );
}

static void dma_outb( var, port )
{
    outb( port, var );
    return;
}

/* enable/disable a specific DMA channel */
static  void enable_dma(unsigned int dmanr)
{
	if (dmanr<=3)
		dma_outb(dmanr,  DMA1_MASK_REG);
	else
		dma_outb(dmanr & 3,  DMA2_MASK_REG);
}

static  void disable_dma(unsigned int dmanr)
{
	if (dmanr<=3)
		dma_outb(dmanr | 4,  DMA1_MASK_REG);
	else
		dma_outb((dmanr & 3) | 4,  DMA2_MASK_REG);
}

/* Clear the 'DMA Pointer Flip Flop'.
 * Write 0 for LSB/MSB, 1 for MSB/LSB access.
 * Use this once to initialize the FF to a known state.
 * After that, keep track of it. :-)
 * --- In order to do that, the DMA routines below should ---
 * --- only be used while interrupts are disabled! ---
 */
static  void clear_dma_ff(unsigned int dmanr)
{
	if (dmanr<=3)
		dma_outb(0,  DMA1_CLEAR_FF_REG);
	else
		dma_outb(0,  DMA2_CLEAR_FF_REG);
}

/* set mode (above) for a specific DMA channel */
static  void set_dma_mode(unsigned int dmanr, char mode)
{
	if (dmanr<=3)
		dma_outb(mode | dmanr,  DMA1_MODE_REG);
	else
		dma_outb(mode | (dmanr&3),  DMA2_MODE_REG);
}

/* Set only the page register bits of the transfer address.
 * This is used for successive transfers when we know the contents of
 * the lower 16 bits of the DMA current address register, but a 64k boundary
 * may have been crossed.
 */
static  void set_dma_page(unsigned int dmanr, char pagenr)
{
	switch(dmanr) {
		case 0:		dma_outb(pagenr, DMA_PAGE_0); 			break;
		case 1:		dma_outb(pagenr, DMA_PAGE_1); 			break;
		case 2:		dma_outb(pagenr, DMA_PAGE_2); 			break;
		case 3:		dma_outb(pagenr, DMA_PAGE_3); 			break;
		case 5:		dma_outb(pagenr & 0xfe, DMA_PAGE_5);	break;
		case 6:		dma_outb(pagenr & 0xfe, DMA_PAGE_6);	break;
		case 7:		dma_outb(pagenr & 0xfe, DMA_PAGE_7);	break;
	}
}


/* Set transfer address & page bits for specific DMA channel.
 * Assumes dma flipflop is clear.
 */
static  void set_dma_addr(unsigned int dmanr, unsigned int a)
{
	set_dma_page(dmanr, a>>16);

	if (dmanr <= 3)  {
	    dma_outb( a & 0xff, ((dmanr&3)<<1) + IO_DMA1_BASE );
        dma_outb( (a>>8) & 0xff, ((dmanr&3)<<1) + IO_DMA1_BASE );
	}  else  {
	    dma_outb( (a>>1) & 0xff, ((dmanr&3)<<2) + IO_DMA2_BASE );
	    dma_outb( (a>>9) & 0xff, ((dmanr&3)<<2) + IO_DMA2_BASE );
	}
}


/* Set transfer size (max 64k for DMA1..3, 128k for DMA5..7) for
 * a specific DMA channel.
 * You must ensure the parameters are valid.
 * NOTE: from a manual: "the number of transfers is one more
 * than the initial word count"! This is taken into account.
 * Assumes dma flip-flop is clear.
 * NOTE 2: "count" represents _bytes_ and must be even for channels 5-7.
 */
static  void set_dma_count(unsigned int dmanr, unsigned int count)
{
        count--;
	if (dmanr <= 3)  {
	    dma_outb( count & 0xff, ((dmanr&3)<<1) + 1 + IO_DMA1_BASE );
	    dma_outb( (count>>8) & 0xff, ((dmanr&3)<<1) + 1 + IO_DMA1_BASE );
        } else {
	    dma_outb( (count>>1) & 0xff, ((dmanr&3)<<2) + 2 + IO_DMA2_BASE );
	    dma_outb( (count>>9) & 0xff, ((dmanr&3)<<2) + 2 + IO_DMA2_BASE );
        }
}


/* Get DMA residue count. After a DMA transfer, this
 * should return zero. Reading this while a DMA transfer is
 * still in progress will return unpredictable results.
 * If called before the channel has been used, it may return 1.
 * Otherwise, it returns the number of _bytes_ left to transfer.
 *
 * Assumes DMA flip-flop is clear.
 */
static  int get_dma_residue(unsigned int dmanr)
{
	unsigned int io_port = (dmanr<=3)? ((dmanr&3)<<1) + 1 + IO_DMA1_BASE
					 : ((dmanr&3)<<2) + 2 + IO_DMA2_BASE;

	/* using short to get 16-bit wrap around */
	unsigned short count;

	count = 1 + dma_inb(io_port);
	count += dma_inb(io_port) << 8;
	
	return (dmanr<=3)? count : (count<<1);
}

/* These are in kernel/dma.c: */
//extern int request_dma(unsigned int dmanr, const char * device_id);	/* reserve a DMA channel */
//extern void free_dma(unsigned int dmanr);	/* release it again */



#ifndef DMA_MODE_AUTOINIT
#define DMA_MODE_AUTOINIT 0x10
#endif

#define NAME		"satlink"
#define CARD_DMA	1
#define CARD_IOBASE	0x300
#define CARD_IOSIZE	0x20
#define BUFFER_SIZE	32768
#define	SLEEP_TIME	1

//
// Some function which I ripped off the orginal satlink DOS program.
// 
#define MAKEWORD(low, high) ((unsigned)(((unsigned char)(low)) | (((unsigned)((unsigned char)(high))) << 8)))
#define MAKELONG(low, high) ((unsigned long)(((unsigned)(low)) | (((unsigned long)((unsigned)(high))) << 16)))

//
// Some needed module global state flags. These will more than likely
// go away when/if we do a multi-open version of this device.
//
static	int    satlink_active      = 0;           /* usage count            */
static	int    satlink_inited      = 0;           /* initialised            */
static	int    dma_inuse           = 0;           /* has DMA started        */

static	satlink_id          id;
static	struct wait_queue   *satlink_wait_queue = NULL;

//
// Ring Buffer maintance variables.
//
static	char   far *dma_buf        = 0;           /* buffer address for DMA */
static	int    dma_buflen          = 0;           /* length of that buffer  */
static	int    usr_readout         = 0;           /* DMA outgoing offset    */

static int
DetectCard(void)
{
	unsigned serial1, serial2;
	//
	// Have _NO_ idea if this is correct or not! I don't
	// have much to go on... If anyone has the correct way
	// of doing this, please let know.
	//
	id.companyid= MAKEWORD(dma_inb(CARD_IOBASE+0x1c),dma_inb(CARD_IOBASE+0x1b));
	id.groupid	= dma_inb(CARD_IOBASE+0x1d);
	id.userid	= MAKEWORD(dma_inb(CARD_IOBASE+0x1f),dma_inb(CARD_IOBASE+0x1e));

	serial1 = MAKEWORD(dma_inb(CARD_IOBASE+0x10),dma_inb(CARD_IOBASE+0x11));
	serial2 = MAKEWORD(dma_inb(CARD_IOBASE+0x12),dma_inb(CARD_IOBASE+0x13));

	id.serialno	= MAKELONG(serial1,serial2);

	return(1);
}

static void
StopDMA()
{
	//
	// When ever we mess with the DMA controller we have to
	// mask the interrupts or very bad things could happen.
	//
	// Lets be very quick about the next thing..
	//
	sti();
		disable_dma(CARD_DMA);
	cli();

	if ( debug )
		printf( "Dev32.satlink: Stopping DMA transfer.\n");

	//
	// Decrement the use count by one.
	//
	dma_inuse--;
}

static void
StartDMA()
{
	//
	// Since we're programming the DMA controller registers
	// we have to mask the IRQ's. We must be quick about this.
	//
	// Look on the bright side, we only do this a few times.
	//
	sti();
		disable_dma(CARD_DMA);
		clear_dma_ff(CARD_DMA);
		set_dma_mode(CARD_DMA,DMA_MODE_READ | DMA_MODE_AUTOINIT);
		set_dma_addr(CARD_DMA,(unsigned int) dma_addr );
		set_dma_count(CARD_DMA,dma_buflen);
	cli();

	if ( debug )
		printf( "Dev32.satlink: Starting DMA transfer.\n");

	//
	// After we program the controller registers and
	// unmasked the IRQ's then we can start the DMA
	// transfer.
	//
	enable_dma(CARD_DMA);

	//
	// Increment it's use count by one.
	//
	dma_inuse++;
}


static int CheckDMA(void)
{
	int	remain1;
	int	remain2;
	int	numread;

	//
	// return the number of bytes that where read since last
	// time we checked. We need to make sure that we get two
	// reads from the dma controller that are the same.
	//
	remain1 = 0;
	remain2 = 1;

	//
	// We _have_ mask the interrupts while we check the dma
	// controller, if we don't an interrupt could disrupt
	// the checking process. 
	//
	// Don't hang around very long, and _don't_ waste any
	// more CPU cycles than we have to!
	//
	// WARNING: If you have a flaky DMA controller, this
	//          _could_ lock up your system, eh sucks to
	//          be you...
	//
	sti();
		while(remain1 != remain2) {
			remain1 = get_dma_residue(CARD_DMA);
			remain2 = get_dma_residue(CARD_DMA);
		}
	cli();

	//
	// Convert the number of bytes remaining to the number of
	// bytes that have been read.
	//
	numread = dma_buflen - remain1;

	//
	// Return this result to the user.
	//

	if ( debug>1 )
		printf("Dev32.satlink: CheckDMA() numread=%d\n", numread );

	return(numread);
}

int satlink_read( char *buf, int count )
{
	int	ValidRead;
	int	ValidSize;
	register int i,j;

	ValidRead = ValidSize = 0;

	//
	// Check the DMA to see how much has been read in. The
	// DMA chip will wrap around and restart when it gets to
	// the end of the buffer. If usr_readout minus ValidRead
	// equals zero, then the buffer is empty.
	//
	ValidRead = CheckDMA();
	if ((ValidRead - usr_readout) == 0) {
		//
		// Now we go into a hard loop checking the
		// DMA control to wait until data bytes have
		// been received. So that we don't lock the
		// machine up, we'll goto to sleep for 1/100
		// of a secound. During which time the buffer
		// at full speed _could_ get up to 328 bytes
		// from satlite receiver.
		//

		// [jcmichot@teaser.fr]
		// Qnx4 Port call satlink_read() every N x 50ms
		// and manage it's own data buffer area, so we
		// do not need to wait here...
		return( 0 );

#if 0
		while((ValidRead - usr_readout) == 0) {
			//
			// Sleep for 1/100 of a secound.
			//
			delay(10);
			//
			// Re-Check the DMA status.
			//
			ValidRead = CheckDMA();
		}
#endif
	}

	//
	// Check to see if the DMA pointer has wrapped
	// around to be behind the readout pointer.
	//
	// From this calculate the valid readout size of
	// the DMA buffer.
	//
	if(ValidRead > usr_readout)
		ValidSize = ValidRead - usr_readout;
	else
		ValidSize = dma_buflen - usr_readout;

	//printf("ValidSize=%d\n", ValidSize );

	//
	// If the ValidSize is bigger than what the
	// user is asking for knock it down a bit. Only
	// give the user what they ask for.
	//
	if (ValidSize > count)
		ValidSize = count;

	//
	// Now move the data out to the user's address
	// space. The next function checks to see if we
	// overwrite anything. Protects, us from our
	// selves.
	//

	if ( debug>1 )
		{
		printf("memcpy( buf, dma_buf+(usr_readout=%5d), ValidSize=%5d )\r",
			usr_readout, ValidSize );
		fflush( stdout );
		}

	// Qnx4, Do not use libc function, few are not DMA safe

	for( j=0, i=usr_readout; i<(usr_readout+ValidSize); i++ )
		buf[j++] = dma_buf[i];

	//
	// Now update the usr_readout point to reflect
	// the new location of the buffer pointer. Check
	// to make sure that we wrapper the counter around
	// to zero when we reach the end.
	//
	usr_readout += ValidSize;
	if(usr_readout == dma_buflen)
		usr_readout = 0;

	//
	// Tell the user how much they got from us.
	//

	return ValidSize;
}

int satlink_init()
{
	int i;

	//
	// Setup up the DMA buffer.
	//
	dma_buflen	= BUFFER_SIZE;
	dma_buf = alloc_dma_buf( dma_buflen );
	if ( dma_buf == (char *) NULL )
		{
		printf("satlink: alloc_dma_buf() return NULL\n");
		exit(0);
		}

	//
	// Probe and claim the region the card is located in
	//
	if (!DetectCard()) {
		return -ENODEV;
	}

	//
	// Init the ring buffer var(s)
	//
	usr_readout = 0;

	//
	// This will should init the card.
	//
	dma_outb( 0x0c, CARD_IOBASE );

	//
	// Start the DMA running so that we don't waste time
	// checking if we started or not.
	//
	StartDMA();

	//
	// Now update the counters
	//
	satlink_active++;

	printf("Qnx4 Satlink 1.0 Driver\n");
	printf( "I/O port 0x%X dma %d\n", CARD_IOBASE, CARD_DMA );

	//
	// This will should init the card.
	//
	dma_outb( 0x0c, CARD_IOBASE );

/*
	printf( "Company %4X Group %2X User %4X\n",id.companyid,
						id.groupid,
						id.userid);
	printf( "Card Serial Number : %8lX\n",id.serialno);
*/

	//
	// Flag to say that the driver has been init'd
	//
	satlink_inited=1;

	//
	// Inform the kernel that all is well.
	//
	return(0);
}

//
// Release allocated resources
//

void free_dma_buf( int val )
{
	if ( dma_inuse )
		StopDMA();

	if ( selector != -1 )
		qnx_segment_free( selector );

	if ( debug )
		printf("\nGot signal (%d)!\n", val );

	exit(0);
}

//
// Alloc 'nbytes' of memory usable by DMA
//
// Define global: dma_addr = physical memory address
//                return a pointer to define "dma_buf"
//

void far *alloc_dma_buf( long nbytes )
{
	static struct _seginfo sbuf;
	char __far *p;

	if ((selector = qnx_segment_alloc_flags(nbytes, 
		_PMF_NOCACHE|_PMF_GLOBAL|_PMF_DMA_SAFE)) == 0
			|| selector == -1)
		return NULL;

	if (qnx_segment_info( 0, 0, selector, &sbuf) == -1)
		return NULL;
	
	dma_addr = sbuf.addr;

	p = MK_FP( selector, 0 );

	return( p );
}
