#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#if !defined(__QNX__)
#include <memory.h>
#else
#include <sys/select.h>
#endif
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>

#include "satcomd.h"

//
//  Application Global variables.
//
ConfigRec	*config	= (ConfigRec *) NULL;
FileHeader	header;

//
// File descriptor for the satellite device driver.
//
int		satlink;

//
// Application configuration defaults.
//
char		*device_name	= SATLINK_DEV;
char		*config_file	= CONFIG_FILE;
int		debug		= 0;

//
// This function will open the satellite device driver. If
// any errors it will log them, then return a false status
// to the caller.
//
int
InitSatComm(char *devname)
{
	if((satlink = open(devname,O_RDONLY)) < 0) {
		//
		// Error opening the satlink device.
		//
		syslog(LOG_ERR,"Failed to open the satellite device driver");
		return(0);
	}
	syslog(LOG_NOTICE,"Openned Satellite device.");
	return(1);
}

//
// This function do any closing of the satellite device driver.
//
void
ShutSatComm()
{
	//
	// Tell the syslog we're closing the satlink driver.
	//
	syslog(LOG_NOTICE,"Closing Satellite device.");
	close(satlink);
}

//
// This function gets called at the beginning of a file transfer.
//
static int
StartFile()
{
	ConfigRec	*tmpconf;

	//
	// See if we have a configuration record to match the incoming
	// file. If we don't then just ignore the incoming file.
	//
	tmpconf = config;
	while(tmpconf) {
		if(tmpconf->rulebase == header.rbase)
			break;
		tmpconf = tmpconf->next;
	}

	//
	// Set the current configuration information to the tmpconf
	// pointer, which could be null, in the case of an unknown file
	// type or it could be the config record for the given file type.
	//
	header.current = tmpconf;

	//
	// Log the fact that we are receiving a file.
	//
	if(debug)
		syslog(LOG_INFO,"Incoming satellite transmittion, %s is %ld bytes long.",header.fname,header.fsize);
	
	//
	// Ok, now depending on configuration, we'll open a file to put this
	// data into.
	//
	if(header.current) {
		//
		// we know about the file, see if we're allowed to receive it
		// or not.
		//
		if(header.current->active) {
			//
			// Build this final pathname from the configuration
			// information.
			//
			sprintf(header.pathname,"%s/%s",header.current->directory,header.fname);

			//
			// now alert the user.
			//
			if(debug)
				syslog(LOG_INFO,"Creating file %s",header.pathname);

			//
			// Then create the file.
			//
			header.outfd = creat(header.pathname,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
			if(header.outfd < 0) {
				//
				// error occurred while opening incoming file.
				//
				syslog(LOG_ERR,"Error %d creating file %s",errno,header.pathname);
				return(0);
			}
		} else {
			//
			// We're not dealing this file type. Log this fact to
			// the system log.
			//
			header.outfd = -1;

			if(debug)
				syslog(LOG_NOTICE,"Reception of this file type %d, has been disabled.",header.rbase);
		}
	} else {
		//
		// The file being transmitted is not a file type that
		// we're geared to deal with, so just ignore it, but log
		// the fact to the syslog.
		//
		if(debug)
			syslog(LOG_NOTICE,"File type %d unknown, ignoring transmittion.",header.rbase);
	}

	//
	// All's well that ends well
	//
	return(1);
}

static int
EndFile()
{
	char	*program;
	pid_t	pid;


	if(header.current) {
		if(header.current->active) {
			//
			// first things first, close the incoming file handle
			// to flush the data.
			//
			close(header.outfd);

			if(header.crc != header.filecrc) {
				//
				// the computed CRC doesn't match the declared
				// CRC in the file header.
				//
				if(debug)
					syslog(LOG_ERR,"File %s was received with an invalid CRC",header.fname);

				program = header.current->badprog;
			} else {
				if(debug)
					syslog(LOG_INFO,"File %s was received correctly",header.fname);

				program = header.current->goodprog;
			}

			//
			// On an debug level greater that 1, tell us what
			// program was used to process the file.
			//
			if(debug>1)
				syslog(LOG_INFO,"Running %s to process the incoming file.",program);

			//
			// Here is where we would run the processing program.
			//
			if((pid = fork()) == 0) {
				//
				// We're the child of the orignal parent.
				// fork once more and then exit.. to leave
				// the child behind on it's own.
				//
				if(fork() != 0)
					exit(0);
				//
				// At this point we are the child's child. The
				// parent doesn't know we're here..
				//
				if(execlp(program,program,header.pathname,0)) {
					//
					// We failed to execute the specified
					// program.
					//
					syslog(LOG_ERR,"Unable to run %s to process file %s",program,header.pathname);
					exit(1);
				}
			} else {
				//
				// On debug level greater than 2, tell us 
				// that we're waiting for a child. 
				//
				if(debug>2)
					syslog(LOG_INFO,"Waiting for child ID %d to die.",pid);

				waitpid(pid,0,0);

				//
				// On debug level greater than 2, tell us 
				// that we're done waiting for a child. 
				//
				if(debug>2)
					syslog(LOG_INFO,"Child Process ID %d is dead, reap for it.",pid);

			}
		}
	}

	return(1);
}

void
HeaderHunt()
{
	unsigned long	SOH;

	if(debug)
		syslog(LOG_INFO,"Looping to look for a Start of Header signature.");

	//
	// Loop until we see the start of header signature.
	// we will always be successful with this, because
	// we will not return until we have found it.
	//
	while(SOH != 0x0000FFFFL)
		read(satlink,&SOH,sizeof(SOH));

	if(debug)
		syslog(LOG_INFO,"Found start of header signature.");

}

int
HeaderRead()
{
	unsigned char	hbuffer[56];
	int		remain;
	int		len;

	//
	// Read the header data in
	//
	remain = sizeof(hbuffer);
	while(remain) {
		len	= read(satlink,hbuffer+(sizeof(hbuffer) - remain),remain);
		if(len < 0) {
			return(0);
		}

		remain -= len;
	}

	//
	// Now decode the header buffer placing this information into
	// the header data structure. If this function returns a false
	// then the header was invalid.
	//

	//
	// Clear out the header, so that we don't have any
	// question as to the validity of this data structure.
	//
	bzero(&header,sizeof(struct FileHeader));

	//
	// Check to see if this header is a
	// valid header.
	//
	if(strcmp(HDR_YAGI(hbuffer),"YAGI")) {
		if(debug)
			syslog(LOG_INFO,"Invalid header signature.");

		return(0);
	}
	
	memcpy(&header.company,	HDR_COMPANY(hbuffer),	sizeof(unsigned short));
	memcpy(&header.group,	HDR_GROUP(hbuffer),	sizeof(unsigned char));
	memcpy(&header.user,	HDR_USER(hbuffer),	sizeof(unsigned short));

	if((header.company != 0xFFFF) && ((header.group) != 0xFF) && ((header.user) != 0xFFFF)) {
		if(debug)
			syslog(LOG_INFO,"Company/Group/User is not broadcast!");

		return(0);
	}

	//
	// If we get this far it means that we can decode
	// the header.
	//
	memcpy(&header.fname,	HDR_FNAME(hbuffer),	13);
	memcpy(&header.ms_time,	HDR_TIME(hbuffer),	sizeof(unsigned short));
	memcpy(&header.ms_date,	HDR_DATE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.crc,	HDR_CRC(hbuffer),	sizeof(unsigned long));

	memcpy(&header.descnum,	HDR_DESCTAB(hbuffer),	sizeof(unsigned short));
	memcpy(&header.decode,	HDR_DECODE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.rbase,	HDR_RULEBASE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.drop,	HDR_DROP(hbuffer),	sizeof(unsigned short));
	memcpy(&header.misc,	HDR_MISC(hbuffer),	sizeof(unsigned short));
	memcpy(&header.type,	HDR_TYPE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.version,	HDR_VERSION(hbuffer),	sizeof(unsigned short));

	memcpy(&header.fsize,	HDR_FSIZE(hbuffer),	sizeof(unsigned long));

	//
	// Return a true value stating that we were successful.
	//
	return(1);
}

int
FileRead()
{
	unsigned char	fbuffer[SAT_MAXREAD];
	unsigned long	fptr;
	unsigned long	len;
	unsigned char	*byte;
	unsigned long	readsiz;
	fd_set		read_set;
	struct timeval	timeout;
	int		retval;

	header.filecrc	= 0xFFFFFFFFL;
	fptr	= 0;

	//
	// Go into a loop and read the data from the satellite driver as it
	// comes in, place it into the file that was openned in the
	// StartFile() function.
	//
	while(fptr < header.fsize) {

		//
		// Setup for the select statement.
		//
		FD_ZERO(&read_set);
		FD_SET(satlink,&read_set);
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;

		if((retval = select(satlink+1,&read_set,0,0,&timeout)) < 0) {
			//
			// Wow! The select had failed! Report this
			// at once to my loyal subjects!
			//
			syslog(LOG_ERR,"Error during the select statement waiting for data.");
			return(0);
		}

		if(retval == 0) {
			//
			// The data stream has stopped during a file download.
			// that means that we over shot the mark. Reset! Get
			// ready for the next file!
			//
			return(1);
		}

		readsiz = ((header.fsize - fptr) > SAT_MAXREAD) ? SAT_MAXREAD : (header.fsize - fptr);

		len = read(satlink,fbuffer,readsiz);
		if(len <= 0) {
			//
			// Houston - We have a problem.
			//
			syslog(LOG_ERR,"Error reading from the satellite device driver.");
			return(0);
		}
		fptr += len;

		//
		// Compute the CRC on the fly.
		//
		for(byte=fbuffer;byte < (fbuffer + len);byte++)
			header.filecrc = UpdCrc32(header.filecrc,*byte);

		//
		// Write the data out to the working file.
		//
		if(header.current) {
			if(header.current->active) {
				if(write(header.outfd,fbuffer,len) != len) {
					//
					// For some reason we we're unable to write
					// data out to the file. return a false status
					// to the caller.
					//
					syslog(LOG_ERR,"Error writing received satellite data to %s",header.fname);
					return(0);
				}
			}
		}
	}

	//
	// Invert the bits in the CRC. So that they will match up with what
	// is in the header.
	//
	header.filecrc = ~header.filecrc;

	//
	// All's well that ends well.
	//
	return(1);
}

void
DoReceive()
{
	//
	// Open and Initialize the satellite device driver.
	//
	syslog(LOG_NOTICE,"Openning satellite device %s",device_name);
	InitSatComm(device_name);

	while(1) {
		//
		// Go into a loop looking for the header signature of a
		// file. If we get it, proceed to read the file into
		// memory. This function will not return until we have
		// a valid Start of Header signature.
		//
		HeaderHunt();

		//
		// Now that we have the start of header signature
		// proceed to read the header and process the incoming
		// file. If this function returns a false value, then
		// the header didn't pass the sanity check restart this
		// while loop to look for another SOH signature.
		//
		if(!HeaderRead())
			continue;

		//
		// The header was good, now we have to call the start of
		// file function to process any incoming files, for example
		// we'll check to see if we want the file, if so, create the
		// file in the correct working directory.
		//
		if(!StartFile()) {
			//
			// An error occurred while openning the file and
			// starting the transfer. We'll need to log this
			// to the syslog and then restart the while loop.
			//
			// SYSLOG
			//
			continue;
		}

		if(!FileRead()) {
			//
			// An error occurred while downloading the file.
			// log this to the syslog, zero the current filecrc
			// and allow the EndFile() function to clean up the
			// mess and run the bad file program.
			//
			header.filecrc = 0;
		}

		//
		// Process any end of file things, like verifing CRC
		// of the file and running any auto processing programs
		// against the received file.
		//
		EndFile();

		//
		// Go around and start the next file.
		//
	}

	//
	// Close down the satellite driver.
	//
	ShutSatComm();
}

int
ReadConfigFile(char *cfile)
{
	ConfigRec	*celement;
	ConfigRec	*tmprec,*tmprec2;
	FILE		*fd;
	char		line[MAX_CONFSZ];
	char		*field;

	//
	// Temporary holding variables.
	//
	unsigned short	rulebase;
	int		active;
	char		*desc;
	char		*directory;
	char		*badprog;
	char		*goodprog;

	//
	// Check to see if we've done this before
	// if so clear out the old config information.
	//

	if(config) {
		if(debug>1)
			syslog(LOG_INFO,"Clearing old config records out.\n");

		tmprec = config;
		while(tmprec) {
			if(tmprec->desc)
				free(tmprec->desc);
			if(tmprec->directory)
				free(tmprec->directory);
			if(tmprec->badprog)
				free(tmprec->badprog);
			if(tmprec->goodprog)
				free(tmprec->goodprog);
			tmprec2	= tmprec;
			tmprec	= tmprec->next;
			free(tmprec2);
		}
		config = (ConfigRec *) NULL;
	}

	//
	// Open the specified configuration file.
	//
	fd = fopen(cfile,"r");
	if(fd == (FILE *) NULL) {
		//
		// Ouch, cann't open the config file. Scream and die.
		//
		syslog(LOG_ERR,"Failed to open configuration file %s",cfile);
		return(0);
	}

	//
	// Read the config file in and process it.
	//
	while(!feof(fd)) {
		if(fgets(line,MAX_CONFSZ,fd) != (char *) NULL) {
			if((line[0] == '#') || (!strlen(line)))
				continue;

			field = strtok(line,":\r\n");
			if(field != (char *) NULL) {
				rulebase = atoi(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				active = atoi(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				desc = strdup(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				directory = strdup(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				badprog = strdup(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				goodprog = strdup(field);
			} else continue;

			celement = (ConfigRec *) malloc(sizeof(struct ConfigRec));
			if(celement == (ConfigRec *) NULL) {
				//
				// We ran out of memory for the ConfigRec
				// entry, this is a fatal error. Scream and
				// die a very bad death.
				//
				syslog(LOG_ERR,"Unable to allocate memory for Config Record.");
				return(0);
			}

			celement->rulebase = rulebase;
			celement->active = active;
			celement->desc = desc;
			celement->directory = directory;
			celement->badprog = badprog;
			celement->goodprog = goodprog;

			celement->next = config;
			config = celement;
		}
	}

	fclose(fd);
	return(1);
}

int
OpenLogs()
{
	//
	// Open the syslog
	//
	openlog("satcomd",0,LOG_DAEMON);

	return(1);
}

void
CloseLogs()
{
	//
	// Be nice and close the syslog connection.
	//
	closelog();
}


int
main(int argc,char *argv[])
{
	char	optc;

	//
	// Process the command line arguments.
	//
	while(1) {
		optc = getopt(argc,argv,"d:f:vV");
		if(optc == -1)
			break;

		switch(optc) {
			case 'd':
				device_name = strdup(optarg);
				break;
			case 'f':
				config_file = strdup(optarg);
				break;
			case 'v':
				debug++;
				break;
			case 'V':
				printf("Satcomd Version 1.1\n");
				break;
			default:
		}
	}

	//
	// Open all the log files.
	//
	if(!OpenLogs()) {
		//
		// We failed to open one or more
		// log files. The OpenLog() function
		// should have screamed which one was at
		// fault.
		//
		exit(1);
	}

	//
	// Title banner.
	//
	syslog(LOG_INFO,"Starting");

	//
	// Open the config file and read it in.
	//
	if(debug)
		syslog(LOG_INFO,"Reading Config File %s",config_file);

	if(!ReadConfigFile(config_file)) {
		//
		// This means that there was an error while reading the
		// configuration file. We'll need to log it to the syslog
		// and exit the program with an error.
		//
		exit(1);
	}

	//
	// Begin the process of receiving the files.
	//
	DoReceive();

	return(0);
}

